#!/usr/bin/perl
#==========================================================================+
#    This source code is public domain, and may be freely distributed,     |
#    modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO      |
#    WARRANTY OF ANY KIND.						   |
#==========================================================================+
#
# FILE:		add_exports.pl
# AUTHOR:	Craig A. James
#		http://www.modgraph.co.uk/ or http://www.modgraph-usa.com
# DESCRIPTION:
#	Swig produces modules with no symbols exported, meaning that
#	every Daylight Toolkit call would have to be prefixed with the
#	package name, e.g.
#
#		$mol = DayPerl2::dt_smilin(...);
#
#	Under normal circumstances this is good because it prevents
#	namespace pollution.  However, Daylight functions already have
#	the dt_ prefix so they're unlikely to collide with any other
#	names, making the module's name, DayPerl2::,  redundant.
#
#	To fix this, we modify the DayPerl2.pm module after swig generates
#	it by adding all of the Daylight functions to its EXPORT
#       statement.
#======================================================================

require 5.004;
use strict qw(vars);
use strict qw(refs);

my $verbose = 0;	# For debugging

my @dayheaders = qw(dt_mdep.h
		    dt_smiles.h
		    dt_depict.h
		    dt_smarts.h
		    dt_finger.h
		    dt_thor.h
		    dt_merlin.h
		    dt_monomer.h
		    dt_progob.h);

#----------------------------------------------------------------------
# Find all of the functions in the Daylight Toolkit.
#----------------------------------------------------------------------

my @exports;
my @constants;
foreach my $header (@dayheaders) {

    my $file = "$ENV{DY_ROOT}/include/$header";
    print "------------------------------ Opening $file\n" if $verbose;
    open(HEADER, "<$file");
    while (<HEADER>) {
	if (m/dt_e_[a-zA-Z0-9_]*\s*\(/) {
	    my $function = $_;
	    $function =~ s/[\n\r]+//;
	    $function =~ s/^.*(dt_e_[a-zA-Z0-9_]+)\s*\(.*$/$1/;
	    $function =~ s/dt_e_/dt_/;
	    print "$function\n" if $verbose;
	    push(@exports, $function);
	}
	if (m/\#define\s*(D[XM]_|TYP_)/) {
	    my $def = $_;
	    $def =~ s/[\n\r]+//;
	    $def =~ s/^.*\#define\s+(\S+)\s+("[^\"]*"|\S+).*$/\$$1 = \"$2\"/;
	    $def =~ s/\"\"/\"/g;
	    $def =~ s/\"\$/\"\\\$/;	# e.g. "$SMI" to "\$SMI"
	    my $defname = (split(" = ", $def))[0];
	    print "$defname: $def\n" if $verbose;
	    push(@exports, $defname);
	    push(@constants, "$def");
	}
    }
    close HEADER;
}
push(@exports,   "\$NULL_OB");
push(@constants, "\$NULL_OB = 0");

#----------------------------------------------------------------------
# Do the PNG drawing library, too.  It's just like the headers, but
# without the dt_e_xxx --> dt_xxx problem.
#----------------------------------------------------------------------

open(HEADER, "<libdl/dl_gd.h");
while (<HEADER>) {
    if (m/dl_[a-zA-Z0-9_]*\s*\(/) {
	my $function = $_;
	$function =~ s/[\n\r]+//;
	$function =~ s/^.*(dl_[a-zA-Z0-9_]+)\s*\(.*$/$1/;
	print "$function\n" if $verbose;
	push(@exports, $function);
    }
}
close HEADER;


#----------------------------------------------------------------------
# Change the @EXPORT statement in the DayPerl2.pm file.
#----------------------------------------------------------------------

# Read the file into an array
open(DAYPERL, "<DayPerl2.pm");
my @dayperl = <DAYPERL>;
close DAYPERL;

# There are some redundant declarations in the Daylight .h files. 
# Use a hash table to prevent writing them twice.
my %outputNames;

# Write the new file
open(DAYPERL, ">DayPerl2.pm");
foreach my $line (@dayperl) {
    if ($line =~ m/\@EXPORT/) {
	print DAYPERL "# These export statements added by add_exports.pl\n";
	print DAYPERL "\@EXPORT = qw(\n";
	foreach my $function (@exports) {
	    if (!$outputNames{$function}) {
		print DAYPERL "\t$function\n";
		$outputNames{$function} = 1;
	    }
	}
	print DAYPERL "\t);\n";
	foreach my $def (@constants) {
	    if (!$outputNames{$def}) {
		print DAYPERL "my $def;\n";
		$outputNames{$def} = 1;
	    }
	}
	# put in a version while we're at it
	print DAYPERL "my \$VERSION = 1.0;\n";
    } 
    else {
	print DAYPERL $line;
    }
}

close DAYPERL;

print "$0: Added \@EXPORTS to DayPerl2.pm\n";
exit 0;
