#!/usr/bin/perl
#======================================================================
#    This source code is public domain, and may be freely distributed,
#    modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
#    WARRANTY OF ANY KIND.
#==========================================================================+
# FILE:		MDL.pm
# AUTHOR:	Craig A. James
# DESCRIPTION:
#	Converts a molecule to a MOL file string.  
#
#	Based on $DY_ROOT/contrib/src/oracle/mdl/smi2moltalk.c
#	
#	smi2mol.c - convert SMI or TDT file to a MOLfile (MDL SD format), 
#	     optionally adding Depict 2D coordinates.
#	
#	Wildcard atoms (*) with ALAB<> labels are interpreted as MDL
#	pseudo-atoms and written to the MDL symbol field.
#	
#	R/S chirality is translated to stereo parity and up/down bonds.
#	Cis/trans information is reflected in generated coordinates.
#	
#	Reference: JCICS #32, 1992, Dalby et al.
#======================================================================

use strict;
use POSIX qw(abs sin cos acos sqrt);
use DayPerl2;

package MDL;


my $PI = 3.14159;

my $natoms;                   #  Number of atoms to molfile          
my $nbonds;                   #  Number of bonds to molfile          


sub smi2mol {
    my ($smi, $coords2D, $coords3D, $alab, $showChirality) = @_;

    my ($smilen, $wegot_2d, $wegot_3d, $wegot_name) = (0,0,0,0);

    my $mol = ::dt_smilin($smi);
    return ($::NULL_OB) if $::NULL_OB == $mol;

    my ($dummy, $name) = split(/ /, $smi);
    $wegot_name = 1 if $name;

    #------------------------------------------------------------
    # Get $natoms, $nbonds
    #------------------------------------------------------------

    $natoms = ::dt_count($mol, $::TYP_ATOM);
    $nbonds = ::dt_count($mol, $::TYP_BOND);

    #------------------------------------------------------------
    # Associate atoms with their index as per original SMILES order.
    # Compute total implicit H-count.  Note that this is needed for
    # the 2D or D3D xyz's to maintain their association.
    # (Not necessary for the bonds.)
    #------------------------------------------------------------

    my $atomstream = ::dt_origstream($mol, $::TYP_ATOM);
    my $atom;
    my $i = 0;
    while ($::NULL_OB != ($atom = ::dt_next($atomstream))) {
	::dt_setinteger($atom, "index", ++$i);
    }

    my $bondstream = ::dt_stream($mol, $::TYP_BOND);
    $i = 0;
    my $bond;
    while ($::NULL_OB != ($bond = ::dt_next($bondstream))) {
	$i++;
	::dt_setinteger($bond, "index", $i);
	my $xatoms = ::dt_stream($bond, $::TYP_ATOM);
	::dt_sethandle($bond, "from", ::dt_next($xatoms));
	::dt_sethandle($bond, "to",   ::dt_next($xatoms));
	::dt_dealloc($xatoms);
    }

    #------------------------------------------------------------
    # Read in data to string s2d.  Replace
    # commas with spaces; find spaces and sscan
    # each real.
    #------------------------------------------------------------

    if ($coords2D) {
	my @coords = split(/,/, $coords2D);
	$wegot_2d = 1;
	::dt_reset($atomstream);
	while ($#coords >= 0 && ($::NULL_OB != ($atom = ::dt_next($atomstream)))) {
	    ::dt_setreal($atom, "x", shift(@coords));
	    ::dt_setreal($atom, "y", shift(@coords));
	}
    }

    #------------------------------------------------------------
    # Read in D3D data to sd3d.
    # Copy only xyz's to sxyzs (between 2nd and
    # 3rd semicolon).  Replace commas w/ spaces.
    # Find spaces and sscan each real.
    #------------------------------------------------------------

    if ($coords3D) {
	my @coords = split(/,/, $coords3D);
	$wegot_3d = 1;
	::dt_reset($atomstream);
	while ($#coords >= 0 && ($::NULL_OB != ($atom = ::dt_next($atomstream)))) {
	    ::dt_setreal($atom, "x", shift(@coords));
	    ::dt_setreal($atom, "y", shift(@coords));
	    ::dt_setreal($atom, "z", shift(@coords));
	}
    }

    #------------------------------------------------------------
    # Look for ALAB which may be MDL pseudo-atom symbols.
    #------------------------------------------------------------

    if ($alab) {
	my @labels = split(/,/, $alab);
	::dt_reset($atomstream);
	while ($#labels >= 0 && ($::NULL_OB != ($atom = ::dt_next($atomstream)))) {
	    ::dt_setstring($atom, "symbol", shift(@labels));
	}
    }

    #------------------------------------------------------------
    # Generate 2D coordinates if needed
    #------------------------------------------------------------

    if (!$wegot_2d && !$wegot_3d) {  
	my $dep = ::dt_alloc_depiction($mol);
	::dt_setisomeric($dep, 1);
	::dt_calcxy($dep);
	::dt_reset($atomstream);
	while ($::NULL_OB != ($atom = ::dt_next($atomstream))) {
	    my ($ok, $x, $y, $z) = ::dt_getcoord($dep, $atom);
	    ::dt_setreal($atom, "x", $x);
	    ::dt_setreal($atom, "y", $y);
	    ::dt_setreal($atom, "z", $z);
	}
	::dt_dealloc($dep);
	$wegot_2d = 1;
    }

    ::dt_dealloc($atomstream);
    ::dt_dealloc($bondstream);

    #------------------------------------------------------------
    # Set the atom parity for each atom.
    #------------------------------------------------------------

    setparity($mol);

    #------------------------------------------------------------
    # Convert tetrahedral chirality to up/down bonds.
    # Does this make sense if wegot_3d?  I think not.
    # Note that setchibonds() may set mol mod_on.
    # Explicit H's may be added and $natoms and $nbonds increased.
    #------------------------------------------------------------
    if ($wegot_2d && $showChirality) {

  	# Set tetrahedral chirality
  	setchibonds($mol);

  	# Convert cis/trans chirality to stereo.
  	setezstereo($mol);
    }

    return ($::NULL_OB) if !::dt_mod_off($mol);

    #------------------------------------------------------------
    # Let's write out the MOL file.
    # First the "header"
    #------------------------------------------------------------

    my @molFile;

    if ($wegot_name) {
	push(@molFile, sprintf("%.80s", $name));
    } else {
	push(@molFile, "");
    }

    push(@molFile, "  SMI2MOL ");

    # Use cansmi for "comment", but truncate to 80 characters
    my $cansmi = ::dt_cansmiles($mol, 1);
    $cansmi = substr($cansmi,0, 80);
    push(@molFile, $cansmi);

    #------------------------------------------------------------
    # Next the counts-line: 
    #      aaabbblllfffcccsssxxxrrrpppiiimmmvvvvvv
    # Note mmm is set to 1 for M END line.  Needs to be increased
    # if M ISO, M CHG lines added ccc is set to 1 = CHIRAL if
    # the absolute chirality is known.  Can only guarantee this
    # if wegot_3d = TRUE.
    #------------------------------------------------------------

    push(@molFile, sprintf("%3d%3d  0  0%3d  0              1 V2000", $natoms, $nbonds, $wegot_3d));

    #------------------------------------------------------------
    # Now the atoms.  
    #------------------------------------------------------------

    $atomstream = ::dt_stream($mol, $::TYP_ATOM);

    while ($::NULL_OB != ($atom = ::dt_next($atomstream))) {
	
	my $sym = "";
	my $mass_diff = 0;
	my $atomText = sprintf("%10.4f%10.4f%10.4f ",
			       ::dt_real($atom, "x"), ::dt_real($atom, "y"), ::dt_real($atom, "z"));

	#------------------------------------------------------------
	# Deuterium, tritium, and mass-difference.
	#------------------------------------------------------------
	     
	my ($atnum, $mass_diff, $weight);
	if (0 < ($weight = ::dt_weight($atom))) {
	    $atnum = ::dt_number($atom);
	    if    ($atnum == 1 && $weight == 2) { $sym = "D"; $mass_diff = 0; }
	    elsif ($atnum == 1 && $weight == 3) { $sym = "T"; $mass_diff = 0; }
	    else {
		$mass_diff = $weight - atnum2mass($atnum);
	    }
	}
	
	#------------------------------------------------------------
	# Check for wildcard atoms which might have
	# MDL symbols from ALAB<>.
	#------------------------------------------------------------

	my $symbol = ::dt_string($atom, "symbol");
	if (0 == ::dt_number($atom) && $symbol) {
	    $sym = ::dt_string($atom, "symbol");
	} else {
	    $sym = ::dt_symbol($atom);
	}
	$atomText .= sprintf("%-3s%2d", $sym, $mass_diff);

	#------------------------------------------------------------
	# Note format: (charge) = 4 - (MDL charge)
	# --> except 0.
	#------------------------------------------------------------

	my $charge = ::dt_charge($atom);
	if (0 != $charge) {
	    $charge = 4 - $charge;
	}
	$atomText .= sprintf("%3d", $charge);
	$atomText .= sprintf("  %1d", ::dt_integer($atom, "parity"));

	#------------------------------------------------------------
	#   jb2314@ggr.co.uk 1-October-1996    *******
	# H count:
	# This line is used for query and means the following
	# 1 = H0  i.e. no H atoms allowed unless explicitly drawn
	# 2 = H1, 3 = H2 ...5 = H4 i.e. Hn means atom must have
	# n or more H's in excess of explicit H's
	#
	# Should be set to zero for normal use.
	#------------------------------------------------------------

	$atomText .= "  0";
	$atomText .= "  0";		# stereo care
	$atomText .= "  0";		# valence

	push(@molFile, $atomText);
    }

    #------------------------------------------------------------
    # Now the bonds.  Note that we use
    # ::dt_bondorder and not ::dt_bondtype, because
    # aromatic bonds are not allowed in the MDL
    # SD format (only in query molfiles).
    #------------------------------------------------------------

    $bondstream = ::dt_stream($mol, $::TYP_BOND);
    while ($::NULL_OB != ($bond = ::dt_next($bondstream))) {
	my $atom1 = ::dt_handle($bond, "from");
	my $atom2 = ::dt_handle($bond, "to");
	my $border = ::dt_bondorder($bond);
	my $a1 = ::dt_integer($atom1, "index");
	my $a2 = ::dt_integer($atom2, "index");
	push(@molFile, sprintf("%3d%3d%3d%3d", $a1, $a2, $border, 
			       ::dt_integer($bond, "stereo")));
    }

    #------------------------------------------------------------
    # No blank lines allowed after M END before DATA
    #------------------------------------------------------------
    push(@molFile, "M  END"); 

    #------------------------------------------------------------
    #  Done with mol.  Find end of TDT or EOF.
    #------------------------------------------------------------
    ::dt_dealloc($mol);
    return \@molFile;
}


#------------------------------------------------------------
# setparity - set the MDL stereo/chiral parity for each atom.
#------------------------------------------------------------

sub setparity {
    my $mol = shift;

    my $atomstream;		#  Atom stream                   
    my $atom;			#  Atom                          
    my $bondstream;		#  Bond stream                   
    my $bond;			#  Bond                          
    my $bseq;			#  Bond sequence                 
    my $aseq;			#  Atom sequence                 
    my $a1;			#  Atom 1                        
    my $a2;			#  Atom 2                        
    my $nconn;
    my $chival;

    $atomstream = ::dt_stream($mol, $::TYP_ATOM);

  ATOM:
    while ($::NULL_OB != ($atom = ::dt_next($atomstream))) {

	$bondstream = ::dt_stream($atom, $::TYP_BOND);
	$nconn = ::dt_count($bondstream, $::TYP_BOND);
	if ($nconn != 3 && $nconn != 4) {
	    ::dt_dealloc($bondstream);
	    next ATOM;
	}

	$bseq = stream2seq($bondstream);
	::dt_append($bseq, ::dt_isohydro()) if ($nconn == 3);

	$chival = ::dt_chival($atom, $bseq);
	if ($::DX_CHI_THCW != $chival && $::DX_CHI_THCCW != $chival) {
	    ::dt_dealloc($bondstream);
	    ::dt_dealloc($bseq);
	    next ATOM;
	}

	$aseq = ::dt_alloc_seq();
	::dt_reset($bondstream);
	while ($::NULL_OB != ($bond = ::dt_next($bondstream))) {
	    ::dt_append($aseq, ::dt_xatom($atom, $bond));
	}

	#------------------------------------------------------------
	# Sort aseq with respect to order in MOLfile.
	#------------------------------------------------------------

	for (my $i = 1; $i < $nconn; ++$i) {
	    ::dt_reset($aseq);
	    $a1 = ::dt_next($aseq);
	    while ($::NULL_OB != ($a2 = ::dt_next($aseq))) {
		if (::dt_integer($a1, "index") > ::dt_integer($a2, "index")) {
		    ::dt_delete($aseq);
		    ::dt_insert($aseq, $a2);
		} else {
		    $a1 = $a2;
		}
	    }
	}

	#------------------------------------------------------------
	# Reassemble bond-sequence in order.
	#------------------------------------------------------------

	::dt_reset($bseq);
	while ($::NULL_OB != ::dt_next($bseq)) {
	    ::dt_delete($bseq);
	}

	::dt_reset($aseq);
	while ($::NULL_OB != ($a1 = ::dt_next($aseq))) {
	    ::dt_append($bseq, ::dt_bond($atom, $a1));
	}
	::dt_append($bseq, ::dt_isohydro()) if ($nconn == 3);

	my $parity = 0;
	my $chival = ::dt_chival($atom, $bseq);
	$parity = 1 if $chival = $::DX_CHI_THCW;
	$parity = 2 if $chival = $::DX_CHI_THCCW;

	::dt_setinteger($atom, "parity", $parity);
	::dt_dealloc($bondstream);
	::dt_dealloc($bseq);
	::dt_dealloc($aseq);
    }
}

#------------------------------------------------------------
# setchibonds - set molfile up/down bonds from SMILES chirality.
#               This is only done if writing 2D coordinates.
#------------------------------------------------------------

sub setchibonds {
    my $mol = shift;

    my $atomstream = ::dt_stream($mol, $::TYP_ATOM);
    my $atomseq = stream2seq($atomstream);
    ::dt_dealloc($atomstream);

    ::dt_mod_on($mol) if !::dt_mod_is_on($mol);

  ATOM:
    while ($::NULL_OB != (my $atom = ::dt_next($atomseq))) {

	my $batoms = ::dt_stream($atom, $::TYP_BOND);
	my $bondseq = stream2seq($batoms);
	::dt_dealloc($batoms);

	#------------------------------------------------------------
	# Must have four bonds or three + one hydrogen.
	# Now (nbs) == 3 or 4 and (hcount) == 0 or 1.
	#------------------------------------------------------------

	my $nbs = ::dt_count($bondseq, $::TYP_BOND);
	my $hcount = ::dt_imp_hcount($atom);
	if ($nbs + $hcount != 4 || $hcount > 1) {
	    ::dt_dealloc($bondseq);
	    next ATOM;
	}
	::dt_append($bondseq, ::dt_isohydro()) if ($hcount);

	#------------------------------------------------------------
	# If not chiral center, go to next atom.
	#------------------------------------------------------------
	my $chival = ::dt_chival($atom, $bondseq);
	if ($::DX_CHI_THCW != $chival && $::DX_CHI_THCCW != $chival) {
	    ::dt_dealloc($bondseq);
	    next ATOM;
	}

	#------------------------------------------------------------
	# Now it's geometry time.  We locate the center atom
	# at (0,0) and translate the other 4 (or 3).  Then
	# we compute the polar angle to sort these atoms
	# (anticlockwise).
	#------------------------------------------------------------

	::dt_reset($bondseq);
	my (@chiinfo, $bond);
	for (my $j = 0; ($j < $nbs) && ($::NULL_OB != ($bond = ::dt_next($bondseq))); ++$j) {
	    $chiinfo[$j]->{bond} = $bond;
	    my $xatom = ::dt_xatom($atom, $bond);
	    $chiinfo[$j]->{atom} = $xatom;
	    $chiinfo[$j]->{x} = ::dt_real($xatom, "x") - ::dt_real($atom, "x");
	    $chiinfo[$j]->{y} = ::dt_real($xatom, "y") - ::dt_real($atom, "y");
	    $chiinfo[$j]->{r} = POSIX::sqrt(($chiinfo[$j]->{x}) * ($chiinfo[$j]->{x}) +
				($chiinfo[$j]->{y}) * ($chiinfo[$j]->{y}));
	    if (POSIX::abs($chiinfo[$j]->{r}) <= 0.000000001) {
		::dt_dealloc($bondseq);
		next ATOM;
	    }
	    $chiinfo[$j]->{theta} = ($chiinfo[$j]->{y} >= 0.0) ?
		(POSIX::acos($chiinfo[$j]->{x} / $chiinfo[$j]->{r})) :
		    (2.0*$PI - POSIX::acos($chiinfo[$j]->{x} / $chiinfo[$j]->{r}));
	}

	#------------------------------------------------------------
	# Bubble sort of ccworder[].  Start with 0123 order.
	#------------------------------------------------------------
	my @ccworder; #  chiinfo indices in ccw order  
	for (my $j = 0; $j < $nbs; ++$j) { $ccworder[$j] = $j; }
	for (my $j = 0; $j < $nbs - 1; ++$j) {
	    for (my $k = 0; $k < $nbs - 1 - $j; ++$k) {
		if ($chiinfo[$ccworder[$k + 1]]->{theta} < $chiinfo[$ccworder[$k]]->{theta}) {
		    my $l = $ccworder[$k];
		    $ccworder[$k] = $ccworder[$k + 1];
		    $ccworder[$k + 1] = $l;
		}
	    }
	}

	#------------------------------------------------------------
	# Next create a new sequence in ccw order.  Determine
	# chival wrt this order, and set bond stereo up/down
	# accordingly (in one of several possible stereo-
	# equivalent arrangements).
	#------------------------------------------------------------
	my $ccwbseq = ::dt_alloc_seq();
	for (my $j = 0; $j < $nbs; ++$j) {
	    ::dt_append($ccwbseq, $chiinfo[$ccworder[$j]]->{bond});
	}
	::dt_append($ccwbseq, ::dt_isohydro()) if ($hcount);
	my $chival = ::dt_chival($atom, $ccwbseq);


	#------------------------------------------------------------
	# Now we have ccw order:
	# 
	#    $chiinfo[$ccworder[0]]->{atom}         $nbs == 4
	#              |
	#              |
	#  [1] --------*-------- [3]
	#              |
	#              |
	#             [2]
	#                          
	#    ...or...
	#
	#    $chiinfo[$ccworder[0]]->{atom}          $nbs == 3
	#              |
	#              |
	#  [1]---------* - - - - - H  <- not-yet explicit H
	#              |                 
	#              |                 
	#             [2]                
	#                          
	# We "average" both the bond lengths and angles of
	# the 1st and 3rd bonds to locate the new explicit H
	# between them.  Then translate back from relative
	# coordinates.
	#------------------------------------------------------------

	if (3 == $nbs) {
	    ++$natoms;
	    my $newatom = ::dt_addatom($mol, $::DX_ATN_H, 0);
	    ::dt_setinteger($newatom, "index", $natoms);
	    $ccworder[3] = 3;
	    $chiinfo[3]->{atom} = $newatom;
	    ++$nbonds;
	    my $newbond = $chiinfo[3]->{bond} = ::dt_addbond($newatom, $atom, 1);
	    ::dt_setimp_hcount($atom, 0);
	    ::dt_setinteger($newbond, "index", $nbonds);

	    $chiinfo[3]->{r} = ($chiinfo[$ccworder[0]]->{r} + $chiinfo[$ccworder[2]]->{r}) / 2.0;
	    my $signed_ang = 
		($chiinfo[$ccworder[2]]->{theta} - 2*$PI + $chiinfo[$ccworder[0]]->{theta})/2.0;
	    $chiinfo[3]->{theta} = $signed_ang > 0 ?  $signed_ang : 2 * $PI + $signed_ang;
	    $chiinfo[3]->{x} = $chiinfo[3]->{r} * POSIX::cos($chiinfo[3]->{theta});
	    ::dt_setreal($newatom, "x", $chiinfo[3]->{x} + ::dt_real($atom, "x"));
	    $chiinfo[3]->{y} = $chiinfo[3]->{r} * POSIX::sin($chiinfo[3]->{theta});
	    ::dt_setreal($newatom, "y", $chiinfo[3]->{y} + ::dt_real($atom, "y"));
	    $nbs = 4;
	}

	#------------------------------------------------------------
	# Reassign same chival with new explicit H.
	#------------------------------------------------------------

	::dt_dealloc($ccwbseq);
	$ccwbseq = ::dt_alloc_seq();
	for (my $j = 0; $j < $nbs; ++$j) {
	    ::dt_append($ccwbseq, $chiinfo[$ccworder[$j]]->{bond});
	}
	::dt_setchival($atom, $ccwbseq, $chival);

	if ($::DX_CHI_THCW == $chival || $::DX_CHI_THCCW == $chival) {
	    ::dt_reset($ccwbseq);
	    while ($::NULL_OB != (my $bond = ::dt_next($ccwbseq))) {
		::dt_sethandle($bond, "from", $atom);
		::dt_sethandle($bond, "to", ::dt_xatom($atom, $bond));
	    }
	    if ($::DX_CHI_THCW == $chival) {
		::dt_setinteger($chiinfo[$ccworder[0]]->{bond}, "stereo", 6);
		::dt_setinteger($chiinfo[$ccworder[1]]->{bond}, "stereo", 1);
		::dt_setinteger($chiinfo[$ccworder[2]]->{bond}, "stereo", 6);
		::dt_setinteger($chiinfo[$ccworder[3]]->{bond}, "stereo", 1);
	    } else {						
		::dt_setinteger($chiinfo[$ccworder[0]]->{bond}, "stereo", 1);
		::dt_setinteger($chiinfo[$ccworder[1]]->{bond}, "stereo", 6);
		::dt_setinteger($chiinfo[$ccworder[2]]->{bond}, "stereo", 1);
		::dt_setinteger($chiinfo[$ccworder[3]]->{bond}, "stereo", 6);
	    }
	}
	::dt_dealloc($ccwbseq);

    }
    ::dt_dealloc($atomseq);
    return (1);
}
 
#------------------------------------------------------------
# setezstereo - set bond stereo from cis/trans chirality.
# 0 = as per 2Ds; 3 = don't care.
#------------------------------------------------------------

sub setezstereo(::dt_Handle $mol) {
    my $mol = shift;

    my ($ob1, $ob2, $tb, $bs1, $bs2, $atom1, $atom2, $batoms, $bond, $bondstream);

    $bondstream = ::dt_stream($mol, $::TYP_BOND);
    while ($::NULL_OB != ($bond = ::dt_next($bondstream))) {
	if (::dt_bondorder($bond) == $::DX_BTY_DOUBLE) {
	    $batoms = ::dt_stream($bond, $::TYP_ATOM);
	    $atom1 = ::dt_next($batoms);
	    $atom2 = ::dt_next($batoms);
	    ::dt_dealloc($batoms);

	    $ob1 = $ob2 = $::NULL_OB;
	    $bs1 = ::dt_stream($atom1, $::TYP_BOND);
	    $bs2 = ::dt_stream($atom2, $::TYP_BOND);
	  BONDS1:
	    while (($tb = ::dt_next($bs1)) != $::NULL_OB) {
		if ($tb != $bond) { $ob1 = $tb; last BONDS1; }
	    }
	  BONDS2:
	    while (($tb = ::dt_next($bs2)) != $::NULL_OB) {
		if ($tb != $bond) { $ob2 = $tb; last BONDS2; }
	    }
	    ::dt_dealloc($bs1);
	    ::dt_dealloc($bs2);
	    if ($ob1 != $::NULL_OB && $ob2 != $::NULL_OB) {
		if (::dt_dbo($bond, $ob1, $ob2) == $::DX_CHI_NO_DBO) {
		    ::dt_setinteger($bond, "stereo", 3);
		} else {
		    ::dt_setinteger($bond, "stereo", 0);
		}
	    }
	}
    }
    ::dt_dealloc($bondstream);
    return (1);
}

#------------------------------------------------------------
#  stream2seq - Copy a stream to a new sequence.
#------------------------------------------------------------

sub stream2seq {
    my $stream = shift;

    my $ob;
    my $seq = ::dt_alloc_seq();
    while ($::NULL_OB != ($ob = ::dt_next($stream))) {
	::dt_append($seq, $ob);
    }
    return $seq;
}

#------------------------------------------------------------
#  atnum2mass - Return mass given atomic number.
#  Returns total number of neutron and protons for given atnum.
#  If naturally occurring, for the most common isotope.
#  If not naturally occurring, for the most stable isotope.
#  Zero (0) if not known - error.
#  Reference: CRC Handbook of Chemistry and Physics, 60th Ed.
#------------------------------------------------------------

my @mass = (
    0,
    1,  4,  7,  9,  11, 12, 14, 16, 19, 20,
    23, 24, 27, 28, 31, 32, 35, 40, 39, 40,
    45, 48, 51, 52, 55, 56, 59, 58, 63, 64,
    69, 74, 75, 80, 79, 84, 85, 88, 89, 90,
    93, 98, 97, 102,103,106,107,114,115,120,
    121,130,127,132,133,138,139,140,141,142,
    145,152,153,158,159,164,165,166,169,174,
    175,180,181,184,187,192,193,195,197,202,
    205,208,209,209,210,222,223,226,227,232,
    231,238,237,244,243,247,247,251,254,257,
    257,255,256,260,262,0 );
 
sub atnum2mass {
    my $atnum = shift;
    return 0 if (($atnum < 1) || ($atnum > 105));
    return ($mass[$atnum]);
}

