#!/usr/bin/perl
#======================================================================
#    This source code is public domain, and may be freely distributed,
#    modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
#    WARRANTY OF ANY KIND.
#======================================================================
# FILE:		merlintest.pl
# AUTHOR:	Craig A. James
# DESCRIPTION:
#	Trivial test program, illustrates and tests the basics of the
#	DayPerl2 Merlin Toolkit interface.
#======================================================================

require 5.004;
use strict 'vars';
use strict 'refs';

use lib '.';
use DayPerl2;

my ($server, $isnew) = dt_mer_server("moonunit2", "merlin", "thor", "");
print "\nGot server:\n";

print "   dt_stringvalue() = '" . dt_stringvalue($server) . "'\n";
print "   dt_typename()    = '" . dt_typename($server) . "'\n";

print "Opening medchem00demo database...";
my ($pool, $isnew) = dt_open($server, "medchem00demo", "r", "");
print "\$pool = $pool, \$isnew = $isnew\n\n";


my $smiftype     = GetFieldType($pool, "\$SMI", 1);
my $simftype     = GetFieldType($pool, "SIMILARITY", 1);
my $nameftype    = GetFieldType($pool, "\$NAM", 1);
my $logpftype    = GetFieldType($pool, "CP", 1);
my $logperrftype = GetFieldType($pool, "CP", 2);
print "$smiftype, $simftype, $nameftype, $logpftype, $logperrftype\n";

my $smicol     = CreateColumn($pool, $smiftype);
my $simcol     = CreateColumn($pool, $simftype);
my $namecol    = CreateColumn($pool, $nameftype);
my $logpcol    = CreateColumn($pool, $logpftype);
my $logperrcol = CreateColumn($pool, $logperrftype);

my $columns = {SMILES     => $smicol,
	       Similarity => $simcol,
	       Name	  => $namecol,
	       CLogP      => $logpcol,
	       CPErr      => $logperrcol};

my $hitlist = dt_mer_alloc_hitlist($pool);	       

PrintRows($pool, $hitlist, $columns);

my ($progress, $status) = dt_mer_similarselect($hitlist,
					       $simcol,
					       $DX_SIMILAR_TANIMOTO, 
					       $DX_ACTION_NEW_LIST,
					       -1,
					       "CN1CCC23C4Oc5c3c(CC1C2C=CC4O)ccc5O",
					       0.8,
					       0.0, 0.0);
$status = ContinueSearch($server, $progress, $status);
my ($progress, $status) = dt_mer_sort($hitlist, $simcol, $DX_SORT_NUM, 0);
				      
PrintRows($pool, $hitlist, $columns);


dt_dealloc($pool);
dt_dealloc($server);

exit 0;

sub PrintRows {
    my ($pool, $hitlist, $columns) = @_;
    print "\n";
    printf("%-20s  %-50s  %-4s  %-5s  %-5s\n",
	   qw(Name Smiles Sim  ClogP CPErr));
    my $nhits = dt_mer_length($hitlist);
    $nhits = 20 if $nhits > 20;

    for (my $row = 0; $row < $nhits; $row++) {
	my $colnum = 0;
	my @values;
	foreach my $colname (qw(Name SMILES Similarity CLogP CPErr)) {
	    $values[$colnum] = dt_mer_cellvalue($columns->{$colname}, $hitlist, $row);
	    $values[$colnum] = "~"  unless $values[$colnum];
	    $colnum++;
	}
	printf("%-20s  %-50s  %4.2f  %5.2f  %-5s\n", (@values));
    }
}

sub GetFieldType {
    my ($poolh, $tag, $fieldnum) = @_;

    # Get the datatype object
    my $dtypeh = dt_getdatatype($poolh, $tag);
    DieOnError("Can't get datatype '$tag'") if $NULL_OB == $dtypeh;

    # Get the Nth field type
    my $ftypeh;
    my $ftype_seqh = dt_stream($dtypeh, $TYP_FIELDTYPE);
    for (my $f = 1; $f <= $fieldnum; $f++) {
	if ($NULL_OB == ($ftypeh  = dt_next($ftype_seqh))) {
	    DieOnError("Can't find field $fieldnum for datatype '$tag'");
	}
    }
    dt_dealloc($ftype_seqh);
    return $ftypeh;
}

sub CreateColumn {
    my ($poolh, $ftypeh) = @_;

    # Create the column for the Nth field
    my $colh = dt_mer_alloc_column($poolh, $ftypeh, $DX_FUNC_FIRST);
    if ($NULL_OB == $colh) {
	DieOnError("Failed to create column of type '" . dt_stringvalue($ftypeh) . "'");
    }
    
    return $colh;
}

sub ContinueSearch {

    my ($server, $progress, $status) = @_;

    my ($done_when, $next_print, $status);
    $next_print = 0;
    $done_when = dt_done_when($server);
    while ($DX_STATUS_IN_PROGRESS == $status) {
	($progress, $status) = dt_continue($server);
	my $percent_done = int(100 * $progress/$done_when);
	if ($percent_done >= $next_print) {
	    printf("%d%% ", $percent_done);
	    $next_print += 5;
	}
    }
    print "\n";
    return $status;
}


sub DieOnError {
    my $msg = shift;
    if ($msg) {
	print "$msg\n";
    } else {
	print "FATAL ERROR:\n";
    }
    my $errs = dt_errors($DX_ERR_WARN);
    my $daylight_error;
    while ($NULL_OB != (my $err = dt_next($errs))) {
	my $errstring = dt_stringvalue($err);
	print $errstring;
	dt_dealloc($err);
    }
    dt_errorclear();
    dt_dealloc($errs);
    exit 1;
}

