#!/usr/bin/perl
#======================================================================
#    This source code is public domain, and may be freely distributed,
#    modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
#    WARRANTY OF ANY KIND.
#======================================================================
# FILE:		react.pl
# AUTHOR:	Craig A. James
# DESCRIPTION:
#	Illustrates using the Daylight Reaction Toolkit with the DayPerl2
#	Perl interface.
#
#	Reads SMILES from STDIN, applies a series of Transformations
#	(generic reactions, specified as SMIRKS), and prints the
#	results on STDOUT.
#
#	Each result is printed as a reaction, but atom-to-atom mapping
#	is suppressed.  Chiral and isotopic information is maintained,
#	and in fact the transformations can specifically alter chirality.
#======================================================================

use strict;
use DayPerl2;

require 'du_mf.lib';
require 'du_mwt.lib';

my @smirksList = 
    ( "[*!O!N!S:1][C:2](=O)[*!O!N!S:3]>>[*!O!N!S:1][C:2](-[*!O!N!S:3])(=NNC(=O)C)",
      "[*!O!N!S:1][C:2](=O)[*!O!N!S:3]>>[*!O!N!S:1][C:2]([H])(-[*!O!N!S:3])(-O)",
      "[*!O!N!S:1][C:2](=O)[*!O!N!S:3]>>[*!O!N!S:1][C:2]1(-[*!O!N!S:3])OCCO1"
      );
		
# Use the Daylight Toolkit to parse the SMIRKS and turn them
# into "transform" objects.

my @transforms;
foreach my $smirks (@smirksList) {
    my $transform = dt_smirkin($smirks);
    die ("invalid SMIRKS: '$smirks'") if ($NULL_OB == $transform);
    push(@transforms, $transform);
}

# LOOP: Read each SMILES from the input line, apply each
# transformation, print the resulting molecules.

while (my $smiles = <STDIN>) {	# read SMILES from standard input

    chomp $smiles;		# get rid of trailing newline
    my $mol = dt_smilin($smiles);
    my $reactantSmiles = dt_cansmiles($mol, 1);

    foreach my $transform (@transforms) {
	my $molCopy = dt_copy($mol);
	my $reactionSeq = dt_utransform($transform, $molCopy, $DX_FORWARD, 0);
	while (my $reaction = dt_next($reactionSeq)) {
	    my $molSeq = dt_stream($reaction, $TYP_MOLECULE);
	    my @productSmiles;
	    while (my $product = dt_next($molSeq)) {
		if (dt_getrole($product, $reaction) == $DX_ROLE_PRODUCT) {
		    my $reactSmiles = dt_cansmiles($product, 1);
		    push(@productSmiles, $reactSmiles);
		}
	    }
	    print $reactantSmiles, ">>" . join(".", @productSmiles), "\n";
	    dt_dealloc($molSeq);
	    dt_dealloc($reaction);
	}
	dt_dealloc($reactionSeq);
	dt_dealloc($molCopy);
    }
    dt_dealloc($mol);
}

exit 0;

