#!/usr/bin/perl
#======================================================================
#    This source code is public domain, and may be freely distributed,
#    modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
#    WARRANTY OF ANY KIND.
#======================================================================
# FILE:		smartsprops.pl
# AUTHOR:	Craig A. James
# DESCRIPTION:
#	Illustrates computing molecular properties with the DayPerl2
#	Perl interface to the Daylight SMILES and SMARTS Toolkits.
#
#	Reads SMILES from STDIN, writes SMILES plus properties on
#	STDOUT in a comma-separated format (CSV) useful for example
#	by Microsoft Excel and other spreadsheet programs.
#======================================================================

use strict;
use DayPerl2;

require 'du_mf.lib';
require 'du_mwt.lib';

my $hbd = dt_smartin("[!#6;!H0]");
my $hba = dt_smartin("[\$([!#6;+0]);!\$([F,Cl,Br,I]);!\$([o,s,nX3]);!\$([Nv5,Pv5,Sv4,Sv6])]");
my $rot = dt_smartin("[!\$(*#*)&!D1]-&!@[!\$(*#*)&!D1]");

die ("invalid HBD SMARTS") if ($NULL_OB == $hbd);
die ("invalid HBA SMARTS") if ($NULL_OB == $hba);
die ("invalid ROT SMARTS") if ($NULL_OB == $rot);

# Print CSV header row
print "SMILES,MWT,MF,NRot,HBD,HBA\n";

while (my $smiles = <>) {	# read SMILES from standard input

    chomp $smiles;		# get rid of trailing newline

    my $mol = dt_smilin($smiles);
    my $mf   = du_mf($mol);
    my $mwt  = du_mwt($mol);
    my $nhbd = number_of_matches($hbd, $mol);
    my $nhba = number_of_matches($hba, $mol);
    my $nrot = number_of_matches($rot, $mol);
    dt_dealloc($mol);

    print "$smiles,$mwt,$mf,$nrot,$nhbd,$nhba\n";
}

exit 0;

#----------------------------------------------------------------------
# Subroutine: number_of_matches
#	Applies the given SMARTS pattern to a molecule and counts the
#	number of times it occurs in the molecule.
#----------------------------------------------------------------------

sub number_of_matches {
    my ($pattern, $molecule) = @_;
    my $pathset = dt_umatch($pattern, $molecule, 0);
    my $stream = dt_stream($pathset, $TYP_PATH);
    my $n = dt_count($stream, $TYP_ANY);
    dt_dealloc($stream);
    dt_dealloc($pathset);
    return ($n >= 0) ? $n : 0;
}

