#!/usr/bin/perl
#======================================================================
# This source code is public domain, and may be freely distributed,
# modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
# WARRANTY OF ANY KIND.
#======================================================================
# FILE:         spresi_reaction_finder.cgi
# AUTHOR:       Craig A. James
# DESCRIPTION:
#       A DayPerl2 example for Merlin: Find the ten most similar
#       compounds in SPRESI (demo) database for which there is a
#       literature citation.
#======================================================================

use strict;
use DayPerl2;

$| = 1;

my ($server, $pool, $hitlist, $isnew);
my ($smiles, $similarity, $smicol, $simcol, $keywdscol);
my ($progress, $status, $nhits, $row);

# Get SMILES from command line
$smiles = $ARGV[0];

($server, $isnew) = dt_mer_server("localhost", "merlin", "thor", "");
($pool,   $isnew) = dt_open($server, "spresi95demo", "r", "");
$hitlist          = dt_mer_alloc_hitlist($pool);               

die_on_error("No hitlist") if ($NULL_OB == $hitlist);

$smicol    = dt_mer_alloc_column($pool, get_field_type($pool, "\$SMI", 1),      $DX_FUNC_FIRST);
$simcol    = dt_mer_alloc_column($pool, get_field_type($pool, "SIMILARITY", 1), $DX_FUNC_FIRST);
$keywdscol = dt_mer_alloc_column($pool, get_field_type($pool, "JA", 5),         $DX_FUNC_FIRST);

die_on_error("Error with columns") if (   $NULL_OB == $smicol
                                       || $NULL_OB == $simcol
                                       || $NULL_OB == $keywdscol);

# Do similarity search
($progress, $status) = dt_mer_similarselect($hitlist, $simcol, $DX_SIMILAR_TANIMOTO,
                                            $DX_ACTION_NEW_LIST, -1, $smiles, 0.2, 0.0, 0.0);
do { ($progress, $status) = dt_continue($server) } while $status == $DX_STATUS_IN_PROGRESS;

# Remove rows that don't have 'reaction' in the Journal keywords column
($progress, $status) = dt_mer_strsearch($hitlist, $keywdscol, $DX_STRING_REGEXP,
                                        $DX_ACTION_DEL_NONHITS, -1, "reaction", undef);
do { ($progress, $status) = dt_continue($server) } while $status == $DX_STATUS_IN_PROGRESS;

# Sort by similarity
($progress, $status) = dt_mer_sort($hitlist, $simcol, $DX_SORT_NUM, 0);
do { ($progress, $status) = dt_continue($server) } while $status == $DX_STATUS_IN_PROGRESS;

$nhits = dt_mer_length($hitlist);
$nhits = 10 if $nhits > 10;

for ($row = 0; $row < $nhits; $row++) {
    $smiles     = dt_mer_cellvalue($smicol,  $hitlist, $row);
    $similarity = dt_mer_cellvalue($simcol,  $hitlist, $row);
    printf "%5.2f $smiles\n", $similarity;
}
exit;

dt_dealloc($pool);
dt_dealloc($server);

exit 0;

#----------------------------------------------------------------------
# Subroutine: get_field_type()
#	Gets the "field type" object for the specified datatype and
#	field number of that datatype.
#----------------------------------------------------------------------

sub get_field_type {
    my ($poolh, $tag, $fieldnum) = @_;

    # Get the datatype object
    my $dtypeh = dt_getdatatype($poolh, $tag);
    die_on_error("Can't get datatype '$tag'") if $NULL_OB == $dtypeh;

    # Get the Nth field type
    my $ftypeh;
    my $ftype_seqh = dt_stream($dtypeh, $TYP_FIELDTYPE);
    for (my $f = 1; $f <= $fieldnum; $f++) {
        if ($NULL_OB == ($ftypeh  = dt_next($ftype_seqh))) {
            die_on_error("Can't find field $fieldnum for datatype '$tag'");
        }
    }
    dt_dealloc($ftype_seqh);
    return $ftypeh;
}

#----------------------------------------------------------------------
# Subroutine: die_on_error()
#	A convenience function: Prints a message, then the Daylight
#	error queue, then quits.
#----------------------------------------------------------------------

sub die_on_error {
    my $msg = shift;
    if ($msg) {
        print "$msg\n";
    } else {
        print "FATAL ERROR:\n";
    }
    my $errs = dt_errors($DX_ERR_WARN);
    my $daylight_error;
    while ($NULL_OB != (my $err = dt_next($errs))) {
        my $errstring = dt_stringvalue($err);
        print $errstring, "\n";
        dt_dealloc($err);
    }
    dt_errorclear();
    dt_dealloc($errs);
    exit 1;
}
