/* -*-C-*-
+======================================================================
|
|    This source code is public domain, and may be freely distributed,
|    modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
|    WARRANTY OF ANY KIND.
|
| FILE:         dl_gd.c
| AUTHOR:       Craig A. James
| DESCRIPTION:
|       Drawing library for Daylight Depict Toolkit that generates PNG
|       files using the gd library.  See http://www.boutell.com/gd/
|       for documentation or the latest version of gd. 
|
|       There are several improvements in this library over older
|       drawing libraries.  Most notably:
|
|         - Use of fonts instead of vector-drawn characters
|         - Anti-aliasing makes nice lines with no "sawtooth" effect
|         - Support of transparent backgrounds (However, note that
|           Internet Exporer does NOT show transparent PNG's correctly,
|           even after six years.  Use Netscape if you want transparent
|           PNG images in a web page.)
|
| The Daylight Depict Toolkit makes calls to the following functions:
|
|    DL_RECT   dl_bounds(void)          { return rect; }
|    void      dl_drawbox(DL_RECT *r)   { }
|    void      dl_drawlabel(DL_LABEL *l){ }
|    void      dl_drawoval(DL_RECT *r)  { }
|    void      dl_drawto(DL_POINT *p)   { }
|    void      dl_fitbounds(DL_RECT *r) { }
|    void      dl_fitlabel(DL_LABEL *l, DL_RECT *box)    { }
|    void      dl_flush(void)           { }
|    DL_RECT   dl_labelbounds(DL_LABEL *l)   { return rect; }
|    void      dl_moveto(DL_POINT *p)   { }
|    float     dl_paspect(void)         { return 1.0; }
|    void      dl_setga(int ga)         { }
|
+======================================================================
*/

#include <dl_defs.h>
#include <gd.h>

#include <gdfontt.h>
#include <gdfonts.h>
#include <gdfontmb.h>
#include <gdfontl.h>
#include <gdfontg.h>

#include "dl_gd.h"

#define DEBUG 0


/*======================================================================
* Structure to hold info about our image.
======================================================================*/

typedef struct {
  gdImagePtr    gdImage;
  DL_POINT      current_pt;
  DL_RECT       bounds;
  DL_PIXEL      size;
  int           ga;
  void         *pngImagePtr;
  int           bgColor;
  int           currentColor;
  DL_POINT      worldScale;
  DL_POINT      worldOffset;
  gdFontPtr     fontStd;
  gdFontPtr     fontSmall;
}  dl_Image;

/*======================================================================
* The image.  Unfortunately, the Daylight Depict Toolkit provides no
* context, so we have to keep a global image.  It's NOT THREAD SAFE!
======================================================================*/

dl_Image theImage;

/*======================================================================
* GA's.  There are 255 possible Graphics Attributes.  Any GA outside
* of the range 0-255 is silently ignored.
======================================================================*/

#define MAX_GA 256
int gaArray[MAX_GA];

/*======================================================================
* Font stuff
======================================================================*/

#define DL_FONT_TINY   0
#define DL_FONT_SMALL  1
#define DL_FONT_MEDIUM 2
#define DL_FONT_LARGE  3
#define DL_FONT_GIANT  4
#define DL_NUM_FONTS   5



/***************************************************************************
* FUNCTION: set_font_size
*
* DESCRIPTION:
*       Picks the best font size for the given boundary.
*
*       Daylight uses a bond length of 1 unit in the world-coordinate system.
*       A font that's about 0.25 units works out best.  This subroutine goes
*       through the fonts and attempts to find whose size, when translated
*       from pixels to world coordinates, is closest to 0.25 units.
***************************************************************************/

static gdFontPtr fontsArray[DL_NUM_FONTS];

static int set_font_size()
{
  float score, bestscore, h;
  int best, i;

  /* First time */
  if (NULL == fontsArray[0]) {
    fontsArray[DL_FONT_TINY]   = gdFontTiny;
    fontsArray[DL_FONT_SMALL]  = gdFontSmall;
    fontsArray[DL_FONT_MEDIUM] = gdFontMediumBold;
    fontsArray[DL_FONT_LARGE]  = gdFontLarge;
    fontsArray[DL_FONT_GIANT]  = gdFontGiant;
  }

  best = 0;
  bestscore = 999.0;
  for (i = 0; i < DL_NUM_FONTS; i++) {
    h = fontsArray[i]->h / theImage.worldScale.y;
    score = fabs(1.0 - h/0.35);
    if (score < bestscore) {
      bestscore = score;
      best = i;
    }
  }

  theImage.fontStd = fontsArray[best];
  if (best > 0) best--;
  theImage.fontSmall = fontsArray[best];
  
#if DEBUG
  fprintf(stderr, "set_font_size: %d/%d pixels high\n", theImage.fontStd->h, theImage.fontSmall->h);
#endif
}


/***************************************************************************
* FUNCTION: dl_alloc_image
*
* DESCRIPTION:
*       Allocates a new image object and sets its background color.
*
* RETURNS: true/false indicating success/failure
***************************************************************************/

int dl_alloc_image(int width, int height, int bgRed, int bgGreen, int bgBlue, int transparent)
{
  theImage.gdImage = gdImageCreateTrueColor(width, height);
  if (NULL == theImage.gdImage) {
    return 0;
  }
  theImage.bounds.xmin = 0.0;
  theImage.bounds.ymin = 0.0;
  theImage.bounds.xmax = 1.0;
  theImage.bounds.ymax = 1.0;
  theImage.size.h = width;
  theImage.size.v = height;
  theImage.pngImagePtr = NULL;

  theImage.bgColor = gdImageColorAllocate(theImage.gdImage, bgRed, bgGreen, bgBlue);
  gdImageFilledRectangle(theImage.gdImage, 0, 0, width, height, theImage.bgColor);

  if (transparent) {
    gdImageColorTransparent(theImage.gdImage, theImage.bgColor);
  }

  theImage.worldScale.x = 100.0;
  theImage.worldScale.y = 100.0;

  set_font_size();
  return 1;
}

/***************************************************************************
* FUNCTION: dl_free_image
*
* DESCRIPTION:
*       Deallocates an image object.
***************************************************************************/

void dl_free_image()
{
  if (NULL != theImage.gdImage) {
    gdImageDestroy(theImage.gdImage);
    theImage.gdImage = NULL;
  }
  if (NULL != theImage.pngImagePtr) {
    gdFree(theImage.pngImagePtr);
    theImage.pngImagePtr = NULL;
  }
}

/***************************************************************************
* FUNCTION: dl_get_image
*
* DESCRIPTION:
*       Gets a pointer to the PNG image raw binary data.
***************************************************************************/

void *dl_get_image(int *plen)
{
  /* If there's an old image there, free it */
  if (NULL != theImage.pngImagePtr) {
    gdFree(theImage.pngImagePtr);
    theImage.pngImagePtr = NULL;
  }
  /* Create the new image binary */
  theImage.pngImagePtr = gdImagePngPtr(theImage.gdImage, plen);
  if (NULL == theImage.pngImagePtr) {
    *plen = -1;
  }
  return theImage.pngImagePtr;
}

/***************************************************************************
* FUNCTION: dl_alloc_ga
*
* DESCRIPTION:
*       Creates a graphics attribute of the given color (RGB).
*
* RETURNS: ()
***************************************************************************/

void dl_alloc_ga(int ga, int r, int g, int b)
{
  if (ga >= MAX_GA || ga < 0) {
    return;
  }
  gaArray[ga] = gdImageColorAllocate(theImage.gdImage, r, g, b);
}

/***************************************************************************
* FUNCTION: dl_fitbounds
*
* DESCRIPTION:
*       Specifies the boundaries of the drawing area, in world coordinates.
*       "World" coordinates are the units used by the Depict Toolkit.
***************************************************************************/

void dl_fitbounds(DL_RECT *r) {

  float xscale, yscale, width, height;
  int pixwidth, pixheight, extra_pix;

#if DEBUG
  fprintf(stderr, "dl_fitbounds: xmin = %4.2f, xmax = %4.2f, ymin = %4.2f, ymax = %4.2f\n",
          r->xmin, r->xmax, r->ymin, r->ymax);
#endif

  theImage.bounds.xmin = r->xmin;
  theImage.bounds.xmax = r->xmax;
  theImage.bounds.ymin = r->ymin;
  theImage.bounds.ymax = r->ymax;

  theImage.worldOffset.x = theImage.bounds.xmin;
  theImage.worldOffset.y = theImage.bounds.ymin;

  /* Figure out scale factors to map world coordinates to pixel coordinates */
  xscale = (theImage.size.h) / (theImage.bounds.xmax - theImage.bounds.xmin);
  yscale = (theImage.size.v) / (theImage.bounds.ymax - theImage.bounds.ymin); 

  /* Keep the image square: Pick the lesser of two and use it for both, and
     center the resulting image. */
  if (xscale < yscale) {
    height = r->ymax - r->ymin;
    pixheight = height * xscale;
    extra_pix = theImage.size.v - pixheight;
    theImage.worldOffset.y -= ((float)extra_pix)/(2.0*xscale);
    yscale = xscale;
  }
  else {
    width = r->xmax - r->xmin;
    pixwidth = width * yscale;
    extra_pix = theImage.size.h - pixwidth;
    theImage.worldOffset.x -= ((float)extra_pix)/(2.0*yscale);
    xscale = yscale;
  }

  theImage.worldScale.x = xscale;
  theImage.worldScale.y = yscale;

  set_font_size();

#if DEBUG
  fprintf(stderr, "scale = (%4.2f, %4.2f), offset = (%4.1f, %4.1f), extra_pix = %d\n",
          theImage.worldScale.x, theImage.worldScale.y, theImage.worldOffset.x, theImage.worldOffset.y, extra_pix);
#endif


}

/***************************************************************************
* FUNCTION: world2pix
*
* DESCRIPTION:
*       Converts a world coordinate to a pixel coordinate.
***************************************************************************/

static DL_PIXEL world2pix(DL_POINT *point)
{
  static DL_PIXEL pixPoint;
  pixPoint.h = (point->x - theImage.worldOffset.x) * theImage.worldScale.x;
  pixPoint.v = (point->y - theImage.worldOffset.y) * theImage.worldScale.y;
  return pixPoint;
}


/***************************************************************************
* FUNCTION: dl_bounds
*
* DESCRIPTION:
*       Returns the boundaries of the drawing area, in world coordinates.
***************************************************************************/

DL_RECT dl_bounds(void) {
#if DEBUG
  fprintf(stderr, "dl_bounds() called\n");
#endif
  return theImage.bounds;
}

/***************************************************************************
* FUNCTION: dl_fitlabel
*
* DESCRIPTION:
*       Draw the specified text inside the specified box.
***************************************************************************/

void dl_fitlabel(DL_LABEL *l, DL_RECT *box) {

}

/***************************************************************************
* FUNCTION: dl_paspect
*
* DESCRIPTION:
*       Returns the aspect ratio of the actual image.
***************************************************************************/

float dl_paspect(void) {
  return ((float)(theImage.size.v))/((float)(theImage.size.h));
}

/***************************************************************************
* FUNCTION: dl_setga
*
* DESCRIPTION:
*       Specifies which "ga" (graphical attribute) to use for subsequent
*       drawing operations.
***************************************************************************/

void dl_setga(int ga) {
  if (0 == ga)
    ga = 1;
  theImage.ga = ga;
  theImage.currentColor = gaArray[ga];

#if DEBUG
  fprintf(stderr, "setga: %d\n", ga);
#endif
}

/***************************************************************************
* FUNCTION: dl_moveto
*
* DESCRIPTION:
*       Moves the current point to the given location.
***************************************************************************/

void dl_moveto(DL_POINT *p) {
  theImage.current_pt.x = p->x;
  theImage.current_pt.y = p->y;

#if DEBUG
  fprintf(stderr, "moveto: (%4.2f, %4.2f)\n", p->x, p->y);
#endif
}

/***************************************************************************
* FUNCTION: dl_drawto
*
* DESCRIPTION:
*       Draws a line from the current point to the  given point.
***************************************************************************/

void dl_drawto(DL_POINT *p) {
  DL_PIXEL from, to;
  
  from = world2pix(&theImage.current_pt);
  to   = world2pix(p);
  gdImageSetAntiAliased(theImage.gdImage, theImage.currentColor);
  gdImageLine(theImage.gdImage, from.h, from.v, to.h, to.v, gdAntiAliased);

  theImage.current_pt.x = p->x;
  theImage.current_pt.y = p->y;

#if DEBUG
  fprintf(stderr, "drawto: (%4.2f, %4.2f)\n", p->x, p->y);
  fprintf(stderr, "   pix: %d, %d, %d, %d\n", from.h, from.v, to.h, to.v);
#endif
}

/***************************************************************************
* FUNCTION: dl_drawbox
*
* DESCRIPTION:
*       Draws a box of the specified size, using the current ga.
***************************************************************************/

void dl_drawbox(DL_RECT *r) {

  DL_PIXEL upperleft, lowerright;
  DL_POINT p;

  /* Convert rectangle in world coordinates to circle in pixel coordinates */
  p.x = r->xmin;
  p.y = r->ymax;
  upperleft = world2pix(&p);
  p.x = r->xmax;
  p.y = r->ymin;
  lowerright = world2pix(&p);

  gdImageSetAntiAliased(theImage.gdImage, theImage.currentColor);
  gdImageRectangle(theImage.gdImage, upperleft.h, upperleft.v, lowerright.h, lowerright.v, gdAntiAliased);

#if DEBUG
  fprintf(stderr, "drawbox: center = (%d, %d) to (%d, %d)\n",
          upperleft.h, upperleft.v, lowerright.h, lowerright.v);
#endif
}

/***************************************************************************
* FUNCTION: dl_drawoval
*
* DESCRIPTION:
*       Draws an oval inscribing the given rectangle using the current ga.
***************************************************************************/

void dl_drawoval(DL_RECT *r) {

  DL_PIXEL lowerleft, upperright;
  DL_POINT p;
  int xcenter, ycenter, width, height;

  /* Convert rectangle in world coordinates to circle in pixel coordinates */
  p.x = r->xmin;
  p.y = r->ymin;
  lowerleft = world2pix(&p);
  p.x = r->xmax;
  p.y = r->ymax;
  upperright = world2pix(&p);
  xcenter = (lowerleft.h + upperright.h)/2;
  ycenter = (lowerleft.v + upperright.v)/2;
  width   = upperright.h - lowerleft.h;
  height  = upperright.v - lowerleft.v;

  /* Draw it */
  gdImageSetAntiAliased(theImage.gdImage, theImage.currentColor);
  gdImageArc(theImage.gdImage, xcenter, ycenter, width, height, 0, 360, gdAntiAliased);

#if DEBUG
  fprintf(stderr, "drawoval: center = (%d, %d), size = (%d, %d)\n", xcenter, ycenter, width, height);
#endif
}

/***************************************************************************
* FUNCTION: dl_labelbounds_pix
*
* DESCRIPTION:
*       Computes a label's size in pixels.  (Trivial version for now,
*       ignores subscripts, assumes fixed-size font.)
*
* RETURNS: ()
***************************************************************************/

static void dl_labelbounds_pix(DL_LABEL *label, int *pxmin, int *pymin, int *pxmax, int*pymax)
{
  *pxmin = 0;
  *pymin = 0;
  *pxmax = strlen(label->text) * theImage.fontStd->w;
  *pymax = theImage.fontStd->h;
}

/***************************************************************************
* FUNCTION: dl_labelbounds
*
* DESCRIPTION:
*       Computes the size of the box into which the specified label will fit.
***************************************************************************/

DL_RECT dl_labelbounds(DL_LABEL *label) {
  int x, y, wpix, hpix;
  float wworld, hworld;
  static DL_RECT labelbounds;

  dl_labelbounds_pix(label, &x, &y, &wpix, &hpix);
  wworld = (float)wpix / theImage.worldScale.x;
  hworld = (float)hpix / theImage.worldScale.y;

  labelbounds.xmin = 0.0;
  labelbounds.ymin = 0.0;
  labelbounds.xmax = wworld;
  labelbounds.ymax = hworld;

  return labelbounds;
}

/***************************************************************************
* FUNCTION: dl_drawlabel
*
* DESCRIPTION:
*       Draws a label at the current location.
***************************************************************************/

void dl_drawlabel(DL_LABEL *label) {

  int xminp, yminp, xmaxp, ymaxp, wp, hp;
  DL_PIXEL label_loc_pix;
  DL_PIXEL labelsize_pix;
  
  /*** Figure out label's width and height in pixels ***/
  dl_labelbounds_pix(label, &xminp, &yminp, &xmaxp, &ymaxp);
  wp = xmaxp - xminp;
  hp = ymaxp - yminp;

  /*** Center label on label's location ***/
  label_loc_pix = world2pix(&(label->loc));
  label_loc_pix.h -= wp/2;
  label_loc_pix.v -= hp/2;

  gdImageString(theImage.gdImage, theImage.fontStd, label_loc_pix.h, label_loc_pix.v,
                (unsigned char *) label->text, theImage.currentColor);

#if DEBUG
  fprintf(stderr, "drawlabel: '%s' (%4.2f, %4.2f)\n",
          label->text, label->loc.x, label->loc.y);
#endif
}


/***************************************************************************
* FUNCTION: dl_flush
*
* DESCRIPTION:
*       No-op.  Doesn't do anything for this library.
***************************************************************************/

void dl_flush()
{
  return;
}

